@@ -2,20 +2,14 @@
 #define ANNOY_H
 #include "utils.h"
 
-// Stolen from RcppAnnoy's annoy.cpp, to get it to compile on Win32.
-#if defined(__MINGW32__)
-#undef Realloc
-#undef Free
-#endif
-
-// define R's REprintf as the 'local' error print method for Annoy
-#define __ERROR_PRINTER_OVERRIDE__  REprintf
-
 // turn off AVX always, to avoid small inconsistencies in distance calculations.
 #define NO_MANUAL_VECTORIZATION 1
 
-#include "annoylib.h"
-#include "kissrandom.h"
+#include "RcppAnnoy.h"
+
+#if RCPPANNOY_VERSION < RcppAnnoyVersion(0,1,17,3)
+  typedef AnnoyIndexSingleThreadedBuildPolicy RcppAnnoyIndexThreadPolicy;
+#endif
 
 template<class Distance>
 class Annoy {
@@ -30,7 +24,7 @@ public:
 
     typedef int32_t Index_t;
     typedef float Data_t;
-    typedef AnnoyIndex<Index_t, Data_t, Distance, Kiss64Random> _index;
+    typedef AnnoyIndex<Index_t, Data_t, Distance, Kiss64Random, RcppAnnoyIndexThreadPolicy> _index;
     const std::vector<Index_t>& get_neighbors () const;
     const std::vector<Data_t>& get_distances () const;
 private:
